
let a = Boggle.parse ()
let b = Boggle.parse ()
let () = assert (Array.length a = Array.length b)
module Board = Board.Make(struct let n = Array.length a end)
let to_char = fun x -> x

(** On peut voir une grille comme un automate non-déterministe (avec la
  * simplification due au fait qu'on n'a pas de boucles dans les états).
  *
  * Plus précisemment, une paire hpath*dico est un état, elle peut faire
  * une transition quand on a une extension du hpath compatible avec
  * le dico, et elle est acceptante quand le dico l'est.
  *
  * Pour tester l'équivalence d'automates non-déterministes on fait une
  * déterminisation à la volée: la déterminisation engendre des états
  * avec plusieurs hpaths possibles, mais un unique dico puisqu'il est
  * fixé par le dico initial et le préfixe déja parcouru.
  *
  * Pour vérifier que deux ensembles de d'états X et Y ont le même langage
  * il suffit de vérifier que
  *   - X contient un état acceptant ssi Y contient un état acceptant
  *   - pour tout a, a^-1 X est équivalent à a^-1 Y. *)

(** Check whether a state is accepting, returning a witness if so. *)
let accepting x w =
  try Some (List.find (fun hp -> Trie.mem "" w) x) with
    | Not_found -> None

(** Successors of an hpath in a given board. *)
let successors board hp =
  List.map
    (fun hp' ->
       let (i,j) = List.hd (Board.path_of_hpath hp') in
       let a = board.(i-1).(j-1) in
         a, hp')
    (Board.hpath_next hp)

(** Successors of a set of hpaths, sorted and grouped by letter. *)
let set_successors board s =
  Utils.group
    (List.sort
       (fun (a,_) (b,_) -> compare a b)
       (List.concat (List.map (fun hp -> successors board hp) s)))

exception Counter_example of [`A|`B] * string

(** Le test d'équivalence prend en argument une TODO list d'équivalences
  * à vérifier. Chaque item est de la forme w,x,y où
  *  - w est un dico représenté comme un trie
  *  - x est un ensemble de hpaths dans la 1ere grille
  *  - y est un ensemble de hpaths dans la 2nde grille. *)
let rec equiv = function
  | [] -> ()
  | (w,x,y)::todo ->
      (** Check immediate acceptance *)
      begin match accepting x w, accepting y w with
        | Some hp, None ->
            let w = Boggle.word_of_path to_char a (Board.path_of_hpath hp) in
              raise (Counter_example (`A,w))
        | None, Some hp ->
            let w = Boggle.word_of_path to_char b (Board.path_of_hpath hp) in
              raise (Counter_example (`B,w))
        | _ -> ()
      end ;
      (* Compute transitions of a set for a given boggle board.
       * Each next_* is a list of (char*hpath list) sorted on the first
       * component. *)
      let next_a = set_successors a x in
      let next_b = set_successors b y in
      let rec new_todo todo = function
        | (c,xa)::next_a, (c',xb)::next_b ->
            begin if c < c' then
              match Trie.step_option c w with
                | Some w' -> new_todo ((w',xa,[])::todo) (next_a, (c',xb)::next_b)
                | None -> new_todo todo (next_a, (c',xb)::next_b)
            else if c' < c then
              match Trie.step_option c' w with
                | Some w' -> new_todo ((w',[],xb)::todo) ((c,xa)::next_a, next_b)
                | None -> new_todo todo ((c,xa)::next_a, next_b)
            else
              match Trie.step_option c w with
                | Some w' -> new_todo ((w',xa,xb)::todo) (next_a, next_b)
                | None -> new_todo todo (next_a, next_b)
            end
        | [], (c',xb)::next_b ->
            begin match Trie.step_option c' w with
              | Some w' -> new_todo ((w',[],xb)::todo) ([], next_b)
              | None -> new_todo todo ([], next_b)
            end
        | (c,xa)::next_a, [] ->
            begin match Trie.step_option c w with
              | Some w' -> new_todo ((w',xa,[])::todo) (next_a, [])
              | None -> new_todo todo (next_a, [])
            end
        | [], [] -> todo
      in
        equiv (new_todo todo (next_a,next_b))

let () =
  let words = Trie.load "words.trie" in
    try
      Format.printf "Board A:\n" ;
      Boggle.print to_char a ;
      Format.printf "Board B:\n" ;
      Boggle.print to_char b ;
      equiv [words,[Board.empty_hpath],[Board.empty_hpath]] ;
      Format.printf "Boards are boggle-equivalent!\n"
    with
      | Counter_example (side,w) ->
          Format.printf
            "Word %S found only in %c.\n"
            w (if side = `A then 'A' else 'B')
