
let boggle = Boggle.parse ()
let words = Trie.load "words.trie"

let to_char x = x

module Board = Board.Make(struct let n = Array.length boggle end)

(** Etat de l'itération: nombre de mots trouvés, score,
  * mots uniques trouvés stockés dans un trie, meilleurs mots. *)
type status = {
  total : int ;
  score : int ;
  found : Trie.t ;
  best_words : (int*(string*Board.path)) list
}

let empty_status = {
  total = 0 ; score = 0 ; found = Trie.empty ; best_words = []
}

(** On parcourt tous les chemins, en profondeur.
  * La fonction prend en arg. status (état courant) et une TODO-list
  * de chemins partiels à explorer plus en avant.
  * Chaque chemin a un dico (trie) associé dans lequel on avance
  * au fur et à mesure que le mot partiel grandit, on abandonne
  * le chemin quand le dico devient vide. *)
let rec find_all status = function
  | [] -> status
  | (trie,hpath)::todo ->
      let status,todo =
        let next = Board.hpath_next hpath in
          List.fold_left
            (fun (status,todo) hpath ->
               let path = Board.path_of_hpath hpath in
               let i,j = List.hd path in
                 try
                   let trie = Trie.step boggle.(i-1).(j-1) trie in
                   let todo = (trie,hpath)::todo in
                     if not (Trie.mem "" trie) then status,todo else
                       let w =
                         Boggle.word_of_path ~to_char boggle path
                       in
                       let score = Boggle.score_of_word w in
                       let status =
                         { status with total = status.total+1 }
                       in
                       let status =
                         if Trie.mem w status.found then status else
                           { status with
                               score = status.score+score ;
                               best_words =
                                 (let compare x y = compare (fst x) (fst y) in
                                    Utils.truncate 10
                                      (Utils.sorted_insert ~compare
                                         (score,(w,path)) status.best_words)) ;
                               found =
                                 (* Format.eprintf "> %s\n" w ; *)
                                 Trie.add w status.found }
                       in
                         status,todo
                 with
                   | Not_found -> status,todo)
            (status,todo)
            next
      in
        find_all status todo

let () =
  Format.printf "Board:\n" ;
  Boggle.print (fun x -> x) boggle ;
  Format.printf "Searching...\n%!" ;
  let status = find_all empty_status [(words,Board.empty_hpath)] in
    Format.printf
      "Found %d words, %d unique words.\n"
      status.total (Trie.size status.found) ;
    Format.printf
      "Score: %d\n"
      status.score ;
    Format.printf "Best words:\n" ;
    List.iter
      (fun (s,(w,p)) ->
         Format.printf "%3d points: %s\n" s w)
      status.best_words
