
(** A trie is a search tree indicating at each level whether it contains
  * the empty word and, and giving the possible first characters of contained
  * words together with sub-tries for the tail of those words. *)
type t = {
  has_empty : bool ;
  table : (char*t) list
}

let rec size t =
  let rec aux size = function
    | t :: l ->
        aux
          (if t.has_empty then size+1 else size)
          (List.map snd t.table @ l)
    | [] -> size
  in
    aux 0 [t]

(** {2 Creation} *)

let empty = { has_empty = false ; table = [] }

let add_empty t = { t with has_empty = true }

let singleton ?(start=0) w =
  let rec aux i t =
    if i < start then t else
      aux (i-1) { has_empty = false ; table = [w.[i],t] }
  in
    aux (String.length w - 1) { has_empty = true ; table = [] }

let add w t =
  let l = String.length w in
  let rec aux i t =
    if i = l then add_empty t else
      let rec insert = function
        | (a,t')::tl when a <= w.[i] ->
            if a = w.[i] then
              (a, aux (i+1) t') :: tl
            else
              (a,t') :: insert tl
        | l ->
            (w.[i], singleton ~start:(i+1) w) :: l
      in
        { t with table = insert t.table }
  in
    aux 0 t

(** {2 Lookup} *)

let mem_empty t = t.has_empty

let step a t = List.assoc a t.table

let step_option a t =
  try Some (step a t) with Not_found -> None

let mem w t =
  let l = String.length w in
  (** [aux i t] checks whether the substring of [w] starting
    * at position [i] is in the trie [t]. *)
  let rec aux i t =
    if i = l then t.has_empty else
      match step_option w.[i] t with
        | None -> false
        | Some t -> aux (i+1) t
  in
    aux 0 t

let next_letters t =
  List.map fst t.table

(** {2 Marshaling} *)

let save t ~filename =
  let o = open_out filename in
    Marshal.to_channel o t [] ;
    close_out o

let load ~filename : t =
  let i = open_in filename in
  let t = Marshal.from_channel i in
    close_in i ;
    t
