
let boggle = Boggle.parse ()

module Board = Board.Make(struct let n = Array.length boggle end)

module Cell = struct

  type t = Char of char | Var of char

  let from_char = function
    | 'a' .. 'z' as x -> Char x
    | x -> Var x

end

let boggle =
  Array.map (fun line -> Array.map Cell.from_char line) boggle

(** Use state for variable assignments. *)

let assignment = Hashtbl.create 17

let to_char = function
  | Cell.Char c -> c
  | Cell.Var x -> Hashtbl.find assignment x

(** Lire une case de la grille, éventuellement en l'instantiant
  * selon une des options passées en argument.
  * Calcul non-déterministe codé par double continuation. *)
let get ~options cell ~succ ~fail =
  match cell with
    | Cell.Char c ->
        if List.mem c options then succ c fail else fail ()
    | Cell.Var x ->
        match
          try Some (Hashtbl.find assignment x) with Not_found -> None
        with
          | Some c -> succ c fail
          | None ->
              Utils.pick options ~fail
                ~succ:(fun c k ->
                   assert (not (Hashtbl.mem assignment x)) ;
                   Hashtbl.add assignment x c ;
                   succ c (fun () -> Hashtbl.remove assignment x ; k ()))

(** On cherche un seul chemin sur lequel un mot de trie peut être trouvé,
  * en augmentant l'instantiation "à la demande".
  * En pratique le trie contiendra toujours un unique mot.
  * La recherche fait du backtracking, codé par double continuations. *)
let rec find_one ~succ ~fail (trie,hpath) =
  let next_pos =
    match Board.path_of_hpath hpath with
      | [] -> Board.all_pos
      | pos::_ -> Board.neighbours pos
  in
    Utils.pick next_pos
      ~succ:(fun (i,j) fail ->
         match Board.hpath_step hpath (i,j) with
           | None -> fail ()
           | Some hpath' ->
               let next_letters = Trie.next_letters trie in
                 get ~options:next_letters boggle.(i-1).(j-1) ~fail
                   ~succ:(fun c fail ->
                      match Trie.step_option c trie with
                        | None -> fail ()
                        | Some trie' ->
                            let fail () =
                              find_one ~succ ~fail (trie',hpath')
                            in
                            if Trie.mem "" trie' then
                              let path' = Board.path_of_hpath hpath' in
                              let w =
                                Boggle.word_of_path ~to_char boggle path'
                              in
                                succ (w,path') fail
                            else
                              fail ()))
      ~fail

(** Affiche caractère ou variable. *)
let to_char_X = function
  | Cell.Char c -> c
  | Cell.Var x -> x

(** Affiche caractère, variable ou instantiation de la variable. *)
let to_char_if = function
  | Cell.Char c -> c
  | Cell.Var x ->
     try Hashtbl.find assignment x with Not_found -> x

let debug = false

let () =
  Format.printf "Board:\n" ;
  Boggle.print to_char_X boggle ;
  let rec search i fail =
    if debug then Format.printf "Searching for %S...\n%!" Sys.argv.(i) ;
    find_one
      (Trie.singleton Sys.argv.(i),Board.empty_hpath)
      ~succ:(fun (w,p) fail ->
               if debug || i+1 = Array.length Sys.argv then begin
                 Format.printf
                   "Found %S at %a.\n"
                   w Board.pp_path p ;
                 Format.printf
                   "Assignment: %a\nBoard:\n" 
                   Utils.pp_assignment (Utils.assignment_to_list assignment) ;
                 Boggle.print to_char_if boggle
               end ;
               if i+1 < Array.length Sys.argv then search (i+1) fail)
      ~fail:(fun () ->
               if debug then
                 Format.printf "No solution for %S.\n" Sys.argv.(i) ;
               fail ())
  in
    search 1 (fun () -> Printf.printf "No solution.\n")
